// StandartPSO_C++AMP.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <iostream>
#include <iomanip>
#include <amp.h>
#include <amp_math.h>


using namespace concurrency; 
using namespace concurrency::precise_math;

// Definition of size of swarm and number of dimensions
static const int TileSize = 50;
static const int TileDimension = 30;

// Pseudorandom number generator
struct random_state
{ 
	unsigned long a; 
	unsigned long b; 
	unsigned long c; 
} 
;

unsigned long random(random_state& r) restrict(amp)
{
	unsigned long old = r.b;
    r.b = r.a * 1103515245 + 12345;
    r.a = (~old ^ (r.b >> 3)) - r.c++;
    return r.b;
}

float random_01(random_state& r) restrict(amp)
{
    return (random(r) & 4294967295) / 4294967295.0f;
}

void seed_random(random_state& r, unsigned long seed) restrict(amp)
{
    r.a = seed;
    r.b = 0;
    r.c = 362436;
}

// Computing of test function
float test_function(float* x, int D, int function, float min, float max) restrict(amp)
{
	// Threatment of particle which leave bounds
	for(int d = 0; d < D; ++d)
	{
		if( (x[d] < min) || (x[d] > max) )
		{
			return FLT_MAX;
		}
	}

	float return_value = 0;
	switch(function)
	{
	//1st de jong
	case 1:
		{
		for(int d=0; d<D; d++)
		{
			return_value = return_value + x[d] * x[d];
		}
		break;
		}
	//2nd De Jong
	case 2:
		{
		for(int d = 0; d < D-1; d++)
		{
			return_value = return_value + (100*(((x[d]*x[d])-x[d+1])*((x[d]*x[d])-x[d+1]))+((1-x[d])*(1-x[d])));
		}
		break;
		}
	//3rd De Jong
	case 3:
		{
		for(int d = 0; d < D; d++)
		{
			return_value = return_value + fabs(x[d]);
		}
		break;
		}
	default:
		{
			return_value = 999;
			break;
		}
	}
	return return_value;
}

// Function to parallel calculation
bool RunPSO(accelerator& device, array_view<float,2>& X, array_view<float,2>& P, array_view<float,2>& V, array_view<float,1>& f, int N, int D, float c1, float c2, float chi, int Iter, float min, float max)
{
	// Try for catch TDR error
	try 
    {
		concurrency::parallel_for_each(f.extent.tile<TileSize>(), [=](tiled_index<TileSize> t_idx) restrict(amp)
		{
			// Local ID
			int local_id = t_idx.local[0];
	
			// Random number generation
			unsigned long seeds = (unsigned long)X(local_id,local_id); //using random numbers in population as seeds
			random_state r = {0,0,0};
			seed_random(r, seeds);

			// Definition of array from shared memory
			tile_static float Xt[TileSize][TileDimension];

			// Copy of values from array X (global memory) to array sX (shared memory)
			for(int d = 0; d < D; d++)
			{
				Xt[local_id][d] = X(local_id, d);
			}
			t_idx.barrier.wait_with_tile_static_memory_fence();

			// Pointer to  the current particle
			float* x = Xt[local_id];

			// PSO algorithm
			for(int k = 0; k < Iter; ++k)
			{
				// Fitness calculation of current particle
				float fitness = test_function(x, D, 3, min, max);
				t_idx.barrier.wait_with_tile_static_memory_fence();

				// Saving of better fitness and better position of the particle
				if(fitness < f(local_id))
				{
					f(local_id) = fitness;
					for(int d = 0; d < D; ++d)
					{
						P(local_id,d) = Xt[local_id][d];
					}
				}

				// Backward particle
				int iB = local_id - 1; 

				// Forward particle
				int iF = local_id + 1; 

				// Watching and correction of backward and forward particles
				if(local_id == 0)
				{
					iB = N - 1; 
					iF = 1; 
				} else if( local_id == (N - 1) )
				{
					iB = N - 2; 
					iF = 0; 
				}

				// Position of best particle
				float igb = local_id;
				
				// Reconfiguration of pointer for best particle if forward or backward particle is better
				if(f(iB) < f(local_id))
				{
					igb = iB;

				} else if(f(iF) < f(local_id))
				{
					igb = iF;
				}

				// Calculation of new parameters of the current particle
				for(int d = 0; d < D; ++d)
				{
					float r1 = random_01(r);
					float r2 = random_01(r);

					float p1 = c1 * r1 * (P(local_id,d) -  Xt[local_id][d]);
					float p2 = c2 * r2 * (P(igb, d) - Xt[local_id][d]);
					V(local_id,d) = chi * (V(local_id,d) + p1 + p2); 
					Xt[local_id][d] = Xt[local_id][d] + V(local_id,d);
				}
				t_idx.barrier.wait_with_tile_static_memory_fence();
			}
		});

		// Wait for all threads
		X.synchronize();
		V.synchronize();
		P.synchronize();
		f.synchronize();

	} catch(concurrency::accelerator_view_removed& ex) 
    { 
        std::cout << "TDR exception received: " << ex.what(); 
        std::cout << "Error code:" << std::hex << ex.get_error_code(); 
        std::cout << "Removed reason:" << std::hex  << ex.get_view_removed_reason(); 
        return false; 
    } 

	return true;
}

int _tmain(int argc, _TCHAR* argv[])
{
	// Initialization

	//Initialization of variables
	int Iter = 300000;					// Number of iterations

	const int N = TileSize;				// Number of particles
	const int D = TileDimension;		// Number of dimensions

	float min = -2.048;					// Bound - minimum allowed value	
	float max = 2.047;					// Bound - maximum allowed value

	float c1 = 2.05;					// Learning factor
	float c2 = 2.05;					// Learning factor
	float lambda = c1 + c2;				// Constant lambda 
	float chi = 2 / ( abs(2 - lambda - sqrt(lambda * lambda - 4 * lambda)) );	// Constant chi

	float* Xh = new float[N * D];		// Particles of the swarm
	float* Ph = new float[N * D];		// Best positions of each particle
	float* Vh = new float[N * D];		// Velocity of particles
	float* fh = new float[N];			// Fitness of particles
	float times[30];					// Array for calculation times
	float bests[30];					// Array for best fitnesses

	// Variables for measurement of calculation time
	clock_t start, diff;

	// 30 measurement
	for(int y = 0; y < 30; y++)
	{
		srand(time(0));

		// Initialization of the swarm
		for(int i = 0; i < N; i++)
		{
			fh[i] = FLT_MAX;
			for(int d = 0; d < D; d++)
			{
				// Iterator
				int id = (d + i * D); 

				// Initialization of the swarm with random pozition and zero velocity
				Xh[id] = min + ( (max - min) * (float)rand()/(float)RAND_MAX); 
				Vh[id] = 0;
			}
		}

		// Definition of arrays in global memory
		extent<2> dataExtent(N, D);

		// X
		array_view<float,2> X(dataExtent, Xh); 

		//P
		array_view<float,2> P(dataExtent, Ph);

		//V
		array_view<float,2> V(dataExtent, Vh); 

		//f
		array_view<float> f(N, fh); 

		bool ret = false;

		// Print all comtatible devices
		std::vector<accelerator> accelerators = accelerator::get_all();
		std::for_each(begin(accelerators), end(accelerators),[=](accelerator acc){ 
			std::wcout << "New accelerator: " << acc.description << std::endl;
			std::wcout << "is_debug = " << acc.is_debug << std::endl;
			std::wcout << "is_emulated = " << acc.is_emulated << std::endl;
			std::wcout << "dedicated_memory = " << acc.dedicated_memory << std::endl;
			std::wcout << "device_path = " << acc.device_path << std::endl;
			std::wcout << "has_display = " << acc.has_display << std::endl;                
			std::wcout << "version = " << (acc.version >> 16) << '.' << (acc.version & 0xFFFF) << std::endl;
		});

		// Select first device
		accelerator::set_default(accelerators[0].device_path);
		accelerator ac(accelerator::default_accelerator);

		// Start of measurement of calculation time
		start = clock();

		// Start of parallel calculation
		ret = RunPSO(ac, X, P, V, f, N, D, c1, c2, chi, Iter, min, max);

		// End of measurement of calculation time
		diff = clock() - start;

		// Printing of calculation time
		int msec = diff * 1000 / CLOCKS_PER_SEC;
		std::cout << "Time taken " << msec/1000 << " seconds and " << msec%1000  << " milliseconds\n" <<std::endl;
		times[y]= (float)msec/1000;

		// Printing of best fitness from calculation
		float bestfit = FLT_MAX; 
		for(int i = 0; i < N; i++)
		{
			if(f[i] < bestfit)
			{
				bestfit = f[i];
			}
		}
		bests[y] = bestfit;
		std::cout << std::setprecision(7) << "\nFitness of best swarm particle: " << bestfit <<std::endl;

	}

	// Calculation and printing of average fitness and calculation time and their standart deviations
	float fitmin = FLT_MAX;
	float fitmax = 0;
	float averagetime = 0;
	float averagefit = 0;
	for(int i = 0; i < 30; i++)
	{
		if(bests[i]<fitmin){
			fitmin = bests[i];
		}
		if(bests[i]>fitmax){
			fitmax - bests[i];
		}
		averagetime += times[i];
		averagefit += bests[i];
	}
	averagetime = averagetime/30;
	averagefit = averagefit/30;
	float devtime = 0;
	float devfit = 0;
	for(int i = 0; i < 30; i++)
	{
		devtime += ((times[i]-averagetime)*(times[i]-averagetime));
		devfit += ((bests[i]-averagefit)*(bests[i]-averagefit));
	}
	devtime = devtime/30;
	devtime = sqrt(devtime);
	devfit = devfit/30;
	devfit = sqrt(devfit);
	printf("Average time: %f\nAverage fitness: %f\n Minimum best fitness: %f\n Maximum best fitness: %f", averagetime,averagefit, fitmin, fitmax);
	printf("Standart deviation of time: %f\nStandart deviation of fitness: %f\n ", devtime,devfit);
	
	// Free memory
	delete[] Xh;	
	delete[] Ph; 
	delete[] Vh; 
	delete[] fh; 

	getchar();
	return 0;
}

