#include <utility>
#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include <stdio.h>
#include <iostream>
#include <string>
#include <iterator>
#include <time.h>
#include <fstream>
#include <iostream>
#include <string>
#include <math.h>

// Pseudorandom number generator
typedef struct { unsigned long a, b, c; } random_state;

__device__ unsigned long random(random_state *r)
{
    unsigned long old = r->b;
    r->b = r->a * 1103515245 + 12345;
    r->a = (~old ^ (r->b >> 3)) - r->c++;
    return r->b;
}

__device__ float random_01(random_state *r)
{
    return (random(r) & 4294967295) / 4294967295.0f;
}

__device__ void seed_random(random_state *r, unsigned long seed)
{
    r->a = seed;
    r->b = 0;
    r->c = 362436;
}

// Computing of test function
__device__ float test_function(float* pos, int D, int function, float min, float max)
{
	// Threatment of particle which leave bounds
	for(int d = 0; d < D; ++d)
	{
		if( (pos[d] < min) || (pos[d] > max) )
		{
			return FLT_MAX;
		}
	}

	float return_value = 0;
	switch(function)
	{
	// 1st de jong
	case 1:
		{
		for(int d=0; d<D; d++)
		{
			return_value = return_value + pos[d] * pos[d];
		}
		break;
		}
	// 2nd De Jong
	case 2:
		{
		for(int d = 0; d < D-1; d++)
		{
			return_value = return_value + (100*(((pos[d]*pos[d])-pos[d+1])*((pos[d]*pos[d])-pos[d+1]))+((1-pos[d])*(1-pos[d])));
		}
		break;
		}
	// 3rd De Jong
	case 3:
		{
		for(int d = 0; d < D; d++)
		{
			return_value = return_value + fabs(pos[d]);
		}
		break;
		}
	default:
		{
			return_value = 999;
			break;
		}
	}
	return return_value;
}

// Treatment of errors during the transition from host to the parallel calculation
inline void checkErr(cudaError_t err, const char * name)
{
    if (err != cudaSuccess)
	{
		std::cerr << "ERROR: " << name << " (" << err << ") - " << cudaGetErrorString(err) << std::endl;
		getchar();
		exit(EXIT_FAILURE);
	} 
}

// Definition of function used to transition from host to the parallel calulation
cudaError_t PSO(float *X, float *P, float *V, float *f, float *parameters);

// Function to parallel calculation
__global__ void pso_kernel(float *X, float *P, float *V, float *f, float* parameters)
{
	// Global ID
	int global_id = threadIdx.x;

	// Calculation variables
	int Iter = (int)parameters[0];		// Number of iterations
	int N = (int)parameters[1];			// Number of particles
	int D = (int)parameters[2];			// Number of dimensions
	float c1 = parameters[3];			// Learning factor
	float c2 = parameters[4];			// Learning factor
	float chi = parameters[5];			// Constant chi
	float min = parameters[6];			// Bound - minimum allowed value
	float max = parameters[7];			// Bound - maximum allowed value

	// Random number generation
	unsigned long seeds = (unsigned long)X[global_id]; //using random numbers in population as seeds
	random_state r = {0,0,0};
	seed_random(&r, seeds);

	// Definition of array from shared memory
	extern __shared__ float sX[];

	// Copy of values from array X (global memory) to array sX (shared memory)
	for(int d = 0; d < D; d++)
	{
		sX[(global_id * D) + d] = X[(global_id * D) + d];
	}

	//__syncthreads();

	// Pointers to  the current particle
	float* x = sX + global_id * D;		// The current coordinate of the current particle in the current dimension
	float* p = P + global_id * D;		// The best value of the current particle
	float* v = V + global_id * D;		// The current speed of the current particle in the current dimension

	// PSO algorithm
	for(int k = 0; k < Iter; k++)
	{
		// Fitness calculation of current particle
		float fitness = test_function(x, D, 3, min, max);

		// Saving of better fitness and better position of the particle
		if(fitness < f[global_id])
		{
			f[global_id] = fitness;
			for(int d = 0; d < D; d++)
			{
				p[d] = x[d];
			}
		}		
		
		// Pointer to the current best value of particle
		float* pg = p;

		// Backward particle
		int iB = global_id - 1;
		
		// Forward particle
		int iF = global_id + 1; 

		// Watching and correction of backward and forward particles
		if(global_id == 0)
		{
			iB = N - 1; 
			iF = 1; 
		} else if( global_id == (N - 1) )
		{
			iB = N - 2; 
			iF = 0; 
		}

		// Reconfiguration of pointer for best particle if forward or backward particle is better
		if(f[iB] < f[global_id])
		{
			pg = P + iB * D;

		} else if(f[iF] < f[global_id])
		{
			pg = P + iF * D;
		}

		// Calculation of new parameters of the current particle
		for(int d = 0; d < D; d++)
		{
			float r1 = random_01(&r);
			float r2 = random_01(&r);

			float p1 = c1 * r1 * (p[d] -  x[d]);
			float p2 = c2 * r2 * (pg[d] - x[d]);
			v[d] = chi * (v[d] + p1 + p2); 
			x[d] = x[d] + v[d];
		}
	}
}

int main()
{
	// Initialization

	//Initialization of variables
	int Iter = 300000;		// Number of iterations

	int N = 50;				// Number of particles
	int D = 30;				// Number of dimensions

	float min = -2.048;		// Bound - minimum allowed value	
	float max = 2.047;		// Bound - maximum allowed value

	float c1 = 2.05;		// Learning factor
	float c2 = 2.05;		// Learning factor
	float lambda = c1 + c2; // Constant lambda 
	float chi = 2 / ( abs(2 - lambda - sqrt(lambda * lambda - 4 * lambda)) );	// Constant chi
	
	float* X = new float[N * D];	// Particles of the swarm
	float* P = new float[N * D];	// Best positions of each particle
	float* V = new float[N * D];	// Velocity of particles
	float* f = new float[N];		// Fitness of particles
	float parameters[] = {Iter, N, D, c1, c2, chi, min, max}; // Auxiliary array for calculation variables
	float times[30];				// Array for calculation times
	float bests[30];				// Array for best fitnesses

	// Fullfilling of array parameters
	parameters[0] = Iter;
	parameters[1] = N;
	parameters[2] = D;
	parameters[3] = c1;
	parameters[4] = c2;
	parameters[5] = chi;
	parameters[6] = min;
	parameters[7] = max;

	// Variables for measurement of calculation time
	clock_t start, diff;

	// 30 measurement
	for(int y = 0; y < 30; y++)
	{
		// Initialization of the swarm
		for(int i = 0; i < N; i++)
		{
			f[i] = FLT_MAX;
			for(int d = 0; d < D; d++)
			{
				// Iterator
				int id = (d + i * D);
				
				// Initialization of the swarm with random pozition and zero velocity
				X[id] = min + ( (max - min) * (float)rand()/(float)RAND_MAX); 
				V[id] = 0;
			}
		}

		// Start of measurement of calculation time
		start = clock();

		// Start of parallel calculation
		cudaError_t err = PSO(X, P, V, f, parameters);

		// End of measurement of calculation time
		diff = clock() - start;

		// Checking for error during parallel calculation
		checkErr(err, "PSO IN CUDA FAILED!"); 

		// Printing of calculation time
		int msec = diff * 1000 / CLOCKS_PER_SEC;
		std::cout << "Time taken " << msec/1000 << " seconds and " << msec%1000  << " milliseconds\n" <<std::endl;
		times[y]= (float)msec/1000;

		// Printing of best fitness from calculation
		float bestfit = FLT_MAX; 
		for(int i = 0; i < N; i++)
		{
			if(f[i] < bestfit)
			{
				bestfit = f[i];
			}
		}
		bests[y] = bestfit;
		printf("\nNejlepsi jedinec ma fitness: %f\n",bestfit);

	}

	// Calculation and printing of average fitness and calculation time and their standart deviations 
	float fitmin = FLT_MAX;
	float fitmax = 0;
	float averagetime = 0;
	float averagefit = 0;
	for(int i = 0; i < 30; i++)
	{
		if(bests[i]<fitmin){
			fitmin = bests[i];
		}
		if(bests[i]>fitmax){
			fitmax - bests[i];
		}
		averagetime += times[i];
		averagefit += bests[i];
	}
	averagetime = averagetime/30;
	averagefit = averagefit/30;
	float devtime = 0;
	float devfit = 0;
	for(int i = 0; i < 30; i++)
	{
		devtime += ((times[i]-averagetime)*(times[i]-averagetime));
		devfit += ((bests[i]-averagefit)*(bests[i]-averagefit));
	}
	devtime = devtime/30;
	devtime = sqrt(devtime);
	devfit = devfit/30;
	devfit = sqrt(devfit);
	printf("Average time: %f\nAverage fitness: %f\n Minimum best fitness: %f\n Maximum best fitness: %f", averagetime,averagefit, fitmin, fitmax);
	printf("Standart deviation of time: %f\nStandart deviation of fitness: %f\n ", devtime,devfit);

	// Free memory
	delete[] X;
	delete[] P;	
	delete[] V;	
	delete[] f;	

	// CudaDeviceReset must be called before exiting in order for profiling and tracing tools such as Nsight and Visual Profiler to show complete traces.
    cudaError_t err = cudaDeviceReset();
	checkErr(err, "cudaDeviceReset failed!"); 

	getchar();

	return EXIT_SUCCESS;
    return 0;
}

// Helper function for using CUDA to calculate PSO in parallel.
cudaError_t PSO(float *X, float *P, float *V, float *f, float *parameters)
{
	// Definition of swarm arrays
	float *X_d = 0;
    float *P_d = 0;
    float *V_d = 0;
	float *f_d = 0;
	float *parameters_d = 0;
	cudaError_t err;

	// Choose which GPU to run on, change this on a multi-GPU system.
	err = cudaSetDevice(0);
	checkErr(err, "cudaSetDevice failed!"); 

	// Allocate GPU buffers for vector (X_d)    .
    err = cudaMalloc((void**)&X_d, parameters[1] * parameters[2] * sizeof(float));
	checkErr(err, "cudaMalloc failed!"); 

	// Allocate GPU buffers for vector (P_d)    .
    err = cudaMalloc((void**)&P_d, parameters[1] * parameters[2] * sizeof(float));
    checkErr(err, "cudaMalloc failed!"); 

	// Allocate GPU buffers for vector (V_d)    .
    err = cudaMalloc((void**)&V_d, parameters[1] * parameters[2] * sizeof(float));
    checkErr(err, "cudaMalloc failed!"); 

	// Allocate GPU buffers for vector (f_d)    .
    err = cudaMalloc((void**)&f_d, parameters[1] * sizeof(float));
    checkErr(err, "cudaMalloc failed!"); 

	// Allocate GPU buffers for vector (f_d)    .
    err = cudaMalloc((void**)&parameters_d, 8 * sizeof(float));
    checkErr(err, "cudaMalloc failed!"); 

	// Copy input vector from host memory to GPU buffer (X_d)
    err = cudaMemcpy(X_d, X, parameters[1] * parameters[2] * sizeof(float), cudaMemcpyHostToDevice);
	checkErr(err, "cudaMemcpy failed!"); 

	// Copy input vector from host memory to GPU buffer (P_d)
	err = cudaMemcpy(P_d, P, parameters[1] * parameters[2] * sizeof(float), cudaMemcpyHostToDevice);
    checkErr(err, "cudaMemcpy failed!"); 

	// Copy input vector from host memory to GPU buffer (V_d)
	err = cudaMemcpy(V_d, V, parameters[1] * parameters[2] * sizeof(float), cudaMemcpyHostToDevice);
    checkErr(err, "cudaMemcpy failed!"); 

	// Copy input vector from host memory to GPU buffer (f_d)
	err = cudaMemcpy(f_d, f, parameters[1] * sizeof(float), cudaMemcpyHostToDevice);
    checkErr(err, "cudaMemcpy failed!"); 

	// Copy input vector from host memory to GPU buffer (parameters_d)
	err = cudaMemcpy(parameters_d, parameters, 8* sizeof(float), cudaMemcpyHostToDevice);
    checkErr(err, "cudaMemcpy failed!"); 

	// Launch a kernel on the GPU with one thread for each particle
    int threads = 144;
	if((int)parameters[1]<144)
	{
		threads = parameters[1];
	}
	int grid = ceil(parameters[1]/144);
	pso_kernel<<< grid, threads, parameters[1] * parameters[2] * sizeof(float)>>>(X_d, P_d, V_d, f_d, parameters_d);
    

    // cudaDeviceSynchronize waits for the kernel to finish, and returns
    // any errors encountered during the launch.
    err = cudaDeviceSynchronize();
	checkErr(err, "cudaDeviceSynchronize failed after launching addKernel!"); 

    // Copy output vector from GPU buffer to host memory.
    err = cudaMemcpy(f, f_d, parameters[1] * sizeof(float), cudaMemcpyDeviceToHost);
	checkErr(err, "cudaMemcpy failed!"); 

	// Free GPU memory
    cudaFree(X_d);
    cudaFree(P_d);
    cudaFree(V_d);
	cudaFree(f_d);
	cudaFree(parameters_d);
    
    return err;
}
