#include <utility>
#define __NO_STD_VECTOR // Use cl::vector instead of STD version
#include <CL/cl.hpp>

#include <cstdio>
#include <cstdlib>
#include <fstream>
#include <iostream>
#include <string>
#include <iterator>
#include <ctime>
#include <math.h>

// Treatment of errors during the transition from host to the parallel calculation
inline void checkErr(cl_int err, const char * name)
{
	if (err != CL_SUCCESS) 
	{
		std::cerr << "ERROR: " << name << " (" << err << ")" << std::endl;
		exit(EXIT_FAILURE);
	}
}

int main( )
{
	// Error variable
	cl_int err;

	// Get Platforms
	cl::vector< cl::Platform > platformList;
	cl::Platform::get(&platformList);

	// Treatment of errors during the transition from host to the parallel calculation
	checkErr(platformList.size() != 0 ? CL_SUCCESS : -1, "cl::Platform::get");

	// Print platform
	std::string platformVendor;
	for(int i = 0; i < platformList.size(); ++i)
	{
		platformList[i].getInfo((cl_platform_info)CL_PLATFORM_VENDOR, &platformVendor);
		std::cout << i << " Platform is by: " << platformVendor << "\n";
	}

	// Select platform 0 (NVIDIA)
	cl::Platform selectedPlatform = platformList[0];

	cl_context_properties cprops[3] = 
	{
		CL_CONTEXT_PLATFORM, 
		(cl_context_properties)(selectedPlatform)(), 
		0
	};
	

	// Create context
	cl::Context context(CL_DEVICE_TYPE_DEFAULT, cprops, NULL, NULL, &err);
	checkErr(err, "Context::Context()"); 

	//Initialization of variables
	int Iter = 300000;			// Number of iterations

	int N = 50;					// Number of particles
	int D = 30;					// Number of dimensions

	float min = -100;			// Bound - minimum allowed value
	float max = 100;			// Bound - maximum allowed value


	float c1 = 2.05;			// Learning factor
	float c2 = 2.05;			// Learning factor
	float lambda = c1 + c2;		// Constant lambda
	float chi = 2 / ( abs(2 - lambda - sqrt(lambda * lambda - 4 * lambda)) );	// Constant chi

	cl_float* X = new cl_float[ N * D ];	// Particles of the swarm
	cl_float* P = new cl_float[ N * D ];	// Best positions of each particle
	cl_float* V = new cl_float[ N * D ];	// Velocity of particles
	cl_float* f = new cl_float[ N ];		// Fitness of particles
	cl_float parameters[] = {Iter, N, D, c1, c2, chi, min, max};	 // Auxiliary array for calculation variables
	float times[30];						// Array for calculation times
	float bests[30];						// Array for best fitnesses

	// Variables for measurement of calculation time
	clock_t start, diff;

	// 30 measurement
	for(int y = 0; y < 30; y++)
	{
		// Initialization of the swarm
		for(int i = 0; i < N; i++)
		{
			f[i] = FLT_MAX;
			for(int d = 0; d < D; d++)
			{
				// Iterator
				int id = (d + i * D); // prepocteny index parameteru s indexem [i,d]

				// Initialization of the swarm with random pozition and zero velocity
				X[id] = min + ( (max - min) * (float)rand()/(float)RAND_MAX); 
				P[id] = X[id];
				V[id] = 0;
			}
		}

		// Allocate GPU buffers for vector (XCL) 
		cl::Buffer XCL(context, CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, sizeof(cl_float) * N * D, X, &err); 
		checkErr(err, "Buffer::Buffer(XCL)");

		// Allocate GPU buffers for vector (PCL) 
		cl::Buffer PCL(context, CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, sizeof(cl_float) * N * D, P, &err); 
		checkErr(err, "Buffer::Buffer(PCL)");
	
		// Allocate GPU buffers for vector (VCL) 
		cl::Buffer VCL(context, CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, sizeof(cl_float) * N * D, V, &err); 
		checkErr(err, "Buffer::Buffer(VCL)");
	
		// Allocate GPU buffers for vector (fCL) 
		cl::Buffer fCL(context, CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR, sizeof(cl_float) * N, f, &err); 
		checkErr(err, "Buffer::Buffer(fCL)");

		// Allocate GPU buffers for vector (parametersCL) 
		cl::Buffer parametersCL(context, CL_MEM_READ_ONLY  | CL_MEM_USE_HOST_PTR, sizeof(cl_float) * 8, parameters, &err); 
		checkErr(err, "Buffer::Buffer(parametersCL)");


		// Get Devices
		cl::vector<cl::Device> deviceList;
		deviceList = context.getInfo<CL_CONTEXT_DEVICES>();
		checkErr(deviceList.size() > 0 ? CL_SUCCESS : -1, "deviceList.size() > 0");

		// Device variables
		std::string deviceName;
		cl_ulong localSize;
		cl_uint maxWorkItemDimensions;
		size_t localMaxWorkGroupSize[3];

		// Print devices and their parameters
		for(int i = 0; i < deviceList.size(); ++i)
		{
			deviceList[i].getInfo((cl_device_info)CL_DEVICE_NAME, &deviceName);
			deviceList[i].getInfo((cl_device_info)CL_DEVICE_LOCAL_MEM_SIZE, &localSize);
			deviceList[i].getInfo((cl_device_info)CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, &maxWorkItemDimensions);
		
			deviceList[i].getInfo((cl_device_info)CL_DEVICE_MAX_WORK_GROUP_SIZE, localMaxWorkGroupSize);
		
			std::cout << i << " Device name is: " << deviceName << " local size: " << localSize;
			std::cout << " maxWorkItemDimensions: " << maxWorkItemDimensions;
		
			std::cout << " max work group size: " << localMaxWorkGroupSize[0] << " " << localMaxWorkGroupSize[1] << " "<< localMaxWorkGroupSize[2]  << "\n";
		
		}

		//Choose device 1 (NVIDIA)
		cl::Device device = deviceList[0];
	
	
		// Get Kernel
		std::string filename = "Kernel.cl";
		std::string oclDebugOptions = "-g -s \"" + filename + "\"";
		oclDebugOptions = "";

		std::ifstream file(filename);
	
		checkErr(file.is_open() ? CL_SUCCESS:-1, "Kernel.cl");
	
		std::string prog( std::istreambuf_iterator<char>(file), (std::istreambuf_iterator<char>() ) );
		cl::Program::Sources source(1, std::make_pair(prog.c_str(), prog.length()+1));
	
		cl::Program program(context, source);
		err = program.build(deviceList, oclDebugOptions.c_str() );
		checkErr(err, "Program::build()");

		cl::Kernel kernel(program, "PSO_OpenCL_Kernel", &err);
		checkErr(err, "Kernel::Kernel()");


		// Set arguments

		// XLC
		err = kernel.setArg(0, XCL);
		checkErr(err, "Kernel::setArg()");

		// PCL
		err = kernel.setArg(1, PCL);
		checkErr(err, "Kernel::setArg()");

		// VCL
		err = kernel.setArg(2, VCL);
		checkErr(err, "Kernel::setArg()");

		// parametersCL
		err = kernel.setArg(3, parametersCL);
		checkErr(err, "Kernel::setArg()");

		// fCL
		err = kernel.setArg(4, fCL);
		checkErr(err, "Kernel::setArg()");

		// Set argument for array X stored in local memory
		int groupSize = N;
		err = kernel.setArg<cl::LocalSpaceArg>(5, cl::__local(D * groupSize * sizeof(cl_float)));
		checkErr(err, "Kernel::setArg()");

		// Create command queue
		cl::CommandQueue queue(context, device, 0, &err);
		checkErr(err, "CommandQueue::CommandQueue()");
	
		// Create event
		cl::Event event;

		// Start of measurement of calculation time
		start = clock();

		// Start of parallel calculation
		err = queue.enqueueNDRangeKernel(kernel, cl::NullRange, cl::NDRange(N), cl::NullRange, NULL, &event);

		// Checking for error during parallel calculation
		checkErr(err, "ComamndQueue::enqueueNDRangeKernel()");

		// Wait for all threads
		event.wait();

		// End of measurement of calculation time
		clock_t diff = clock() - start;

		// Printing of calculation time
		int msec = diff * 1000 / CLOCKS_PER_SEC;
		std::cout << "Time taken: " << msec/1000 << " seconds and " << msec%1000 <<" milliseconds"<< std::endl;
		times[y]= (float)msec/1000;

		// Load of argument f from GPU back to host
		err = queue.enqueueReadBuffer(fCL,CL_TRUE,0, sizeof(cl_float) * N, f);
		checkErr(err, "ComamndQueue::enqueueReadBuffer()");

		// Printing of best fitness from calculation
		float bestfit = FLT_MAX; 
		for(int i = 0; i < N; i++)
		{
			if(f[i] < bestfit)
			{
				bestfit = f[i];
			}
		}
		bests[y] = bestfit;
		printf("\nNejlepsi jedinec ma fitness: %f\n",bestfit);

	}

	// Calculation and printing of average fitness and calculation time and their standart deviations 
	float fitmin = FLT_MAX;
	float fitmax = 0;
	float averagetime = 0;
	float averagefit = 0;
	for(int i = 0; i < 30; i++)
	{
		if(bests[i]<fitmin){
			fitmin = bests[i];
		}
		if(bests[i]>fitmax){
			fitmax - bests[i];
		}
		averagetime += times[i];
		averagefit += bests[i];
	}
	averagetime = averagetime/30;
	averagefit = averagefit/30;
	float devtime = 0;
	float devfit = 0;
	for(int i = 0; i < 30; i++)
	{
		devtime += ((times[i]-averagetime)*(times[i]-averagetime));
		devfit += ((bests[i]-averagefit)*(bests[i]-averagefit));
	}
	devtime = devtime/30;
	devtime = sqrt(devtime);
	devfit = devfit/30;
	devfit = sqrt(devfit);
	printf("Average time: %f\nAverage fitness: %f\n Minimum best fitness: %f\n Maximum best fitness: %f", averagetime,averagefit, fitmin, fitmax);
	printf("Standart deviation of time: %f\nStandart deviation of fitness: %f\n ", devtime,devfit);

	getchar();

	// Free memory
	delete[] X;
	delete[] P;
	delete[] V;
	delete[] f;

	return EXIT_SUCCESS;

    return 0;
}